open Std_internal

let%test_module "random" = (module struct

  module G = Quickcheck.Generator
  module O = Quickcheck.Observer

  module type T = sig

    type t [@@deriving sexp_of]

    include Quickcheckable with type t := t
    include Comparable     with type t := t

    val module_name : string

  end

  module type Math = sig

    type t

    include T with type t := t
    include Commutative_group.S with type t := t

  end

  module Make (T : T) (Math : Math) = struct

    open T
    open G.Monad_infix

    module Q = Quickcheck.Configure (struct
        include Quickcheck
        let default_seed =
          `Deterministic
            (sprintf "%s values with %s operators."
               (String.capitalize T.module_name)
               (String.capitalize Math.module_name))
      end)
    ;;

    let%test_unit "mem true" =
      Q.test ~sexp_of:[%sexp_of: t * t list]
        (List.gen gen >>= fun list        ->
         G.of_list                 list  >>| fun elt_of_list ->
         elt_of_list, list)
        ~f:(fun (elt_of_list, list) ->
          [%test_result: bool]
            (List.mem list elt_of_list ~equal)
            ~expect:true)

    let%test_unit "mem false" =
      Q.test ~sexp_of:[%sexp_of: t * t list]
        (G.bind_choice gen (fun choice ->
           let x = G.Choice.value choice in
           let not_x_gen =
             G.Choice.updated_gen choice
               ~keep:`All_choices_except_this_choice
           in
           List.gen not_x_gen
           >>| fun list_of_not_x ->
           x, list_of_not_x))
        ~f:(fun (x, list_of_not_x) ->
          [%test_result: bool]
            (List.mem list_of_not_x x ~equal)
            ~expect:false)

    let%test_unit "len" =
      Q.test ~sexp_of:[%sexp_of: int * t list]
        (G.size
         >>= fun len  ->
         List.gen' gen ~length:(`Exactly len)
         >>| fun list ->
         len, list)
        ~f:(fun (len, list) ->
          [%test_result: int]
            (List.length list)
            ~expect:len)

    let%test_unit "is_empty true" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (G.singleton [])
        ~f:(fun empty ->
          [%test_result: bool]
            (List.is_empty empty)
            ~expect:true)

    let%test_unit "is_empty false" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (G.tuple2 gen (List.gen gen)
         >>| fun (x,list) ->
         x::list)
        ~f:(fun non_empty ->
          [%test_result: bool]
            (List.is_empty non_empty)
            ~expect:false)

    let%test_unit "iter" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          let q = Queue.create () in
          List.iter list ~f:(Queue.enqueue q);
          [%test_result: t list]
            (Queue.to_list q)
            ~expect:list)

    let%test_unit "sum vs fold" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> Math.t)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Math.gen))
        ~f:(fun (list, f) ->
          [%test_eq: Math.t]
            (List.fold list ~init:Math.zero ~f:(fun m x ->
               Math.(+) m (f x)))
            (List.sum (module Math) list ~f))

    let%test_unit "for_all vs exists" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: bool]
            (List.for_all list ~f)
            (not (List.exists list ~f:(Fn.non f))))

    let%test_unit "exists vs mem" =
      Q.test ~sexp_of:[%sexp_of: t * t list]
        (G.tuple2 gen (List.gen gen))
        ~f:(fun (x, list) ->
          [%test_eq: bool]
            (List.exists list ~f:(fun y -> equal x y))
            (List.mem    list x ~equal))

    let%test_unit "exists vs find" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: bool]
            (List.exists list ~f)
            (Option.is_some (List.find list ~f)))

    let%test_unit "count vs length/filter" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: int]
            (List.count list ~f)
            (List.length (List.filter list ~f)))

    let%test_unit "find vs find_map" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t option]
            (List.find     list ~f)
            (List.find_map list ~f:(fun x -> if f x then Some x else None)))

    let%test_unit "to_list" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (List.to_list list)
            ~expect:list)

    let%test_unit "to_array + Array.to_list" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (Array.to_list (List.to_array list))
            ~expect:list)

    let%test_unit "max_elt vs min_elt" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t -> int)]
        (G.tuple2
           (List.gen gen)
           (G.compare_fn obs))
        ~f:(fun (list, cmp) ->
          [%test_eq: t option]
            (List.min_elt list ~cmp)
            (List.max_elt list ~cmp:(fun x y -> cmp y x)))

    let%test_unit "return" =
      Q.test ~sexp_of:[%sexp_of: t]
        gen
        ~f:(fun x ->
          [%test_result: t list]
            (List.return x)
            ~expect:[x])

    let%test_unit "map vs bind" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.map list ~f)
            (List.bind list (fun x -> [ f x ])))

    let%test_unit "monad left identity" =
      Q.test ~sexp_of:[%sexp_of: t * (t -> t list)]
        (G.tuple2 gen (G.fn obs (List.gen gen)))
        ~f:(fun (x, f) ->
          [%test_eq: t list]
            (List.bind (List.return x) f)
            (f x))

    let%test_unit "monad right identity" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (List.bind list List.return)
            ~expect:list)

    let%test_unit "monad associativity" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t list) * (t -> t list)]
        (G.tuple3
           (List.gen gen)
           (G.fn obs (List.gen gen))
           (G.fn obs (List.gen gen)))
        ~f:(fun (list, f, g) ->
          [%test_eq: t list]
            (List.bind (List.bind list f) g)
            (List.bind list (fun x -> List.bind (f x) g)))

    let%test_unit "join" =
      Q.test ~sexp_of:[%sexp_of: t list list]
        (List.gen (List.gen gen))
        ~f:(fun list ->
          [%test_eq: t list]
            (List.join list)
            (List.bind list Fn.id))

    let%test_unit "ignore" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_eq: unit list]
            (List.ignore_m list)
            (List.map list ~f:ignore))

    let%test_unit "of_list + to_list" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (List.of_list list)
            ~expect:list)

    let%test_unit "nth vs nth_exn" =
      Q.test ~sexp_of:[%sexp_of: int * t list]
        (G.tuple2 G.size (List.gen gen))
        ~f:(fun (i, list) ->
          [%test_eq: t option]
            (List.nth list i)
            (Option.try_with (fun () ->
               List.nth_exn list i)))

    let%test_unit "init + nth_exn" =
      Q.test ~sexp_of:[%sexp_of: int * int * (int -> t)]
        (G.size >>= fun size ->
         let size = size + 1 in
         G.tuple3
           (G.return size)
           (Int.gen_between
              ~lower_bound:(Incl 0)
              ~upper_bound:(Excl size))
           (G.fn Int.obs gen))
        ~f:(fun (size, i, f) ->
          [%test_result: t]
            (List.nth_exn (List.init size ~f) i)
            ~expect:(f i))

    let%test_unit "rev^2" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (List.rev (List.rev list))
            ~expect:list)

    let%test_unit "rev_append vs rev + append" =
      Q.test ~sexp_of:[%sexp_of: t list * t list]
        (G.tuple2
           (List.gen gen)
           (List.gen gen))
        ~f:(fun (list1, list2) ->
          [%test_eq: t list]
            (List.rev_append list1 list2)
            (List.append (List.rev list1) list2))

    let%test_unit "unordered_append vs append" =
      Q.test ~sexp_of:[%sexp_of: t list * t list * t]
        (G.tuple3
           (List.gen gen)
           (List.gen gen)
           gen)
        ~f:(fun (list1, list2, x) ->
          [%test_eq: bool]
            (List.mem (List.append list1 list2) x ~equal)
            (List.mem (List.unordered_append list1 list2) x ~equal))

    let%test_unit "rev_map vs map + rev" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.rev_map list ~f)
            (List.rev (List.map list ~f)))

    let%test_unit "fold vs fold_left" =
      Q.test ~sexp_of:[%sexp_of: t list * t * (t -> t -> t)]
        (G.tuple3
           (List.gen gen)
           gen
           (G.fn2 obs obs gen))
        ~f:(fun (list, init, f) ->
          [%test_eq: t]
            (List.fold list ~init ~f)
            (List.fold_left list ~init ~f))

    let%test_unit "unzip + iter2_exn vs iter" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * (t -> t -> t)]
        (G.tuple2
           (List.gen (G.tuple2 gen gen))
           (G.fn2 obs obs gen))
        ~f:(fun (pair_list, f) ->
          [%test_eq: t list]
            (let q = Queue.create () in
             let list1, list2 = List.unzip pair_list in
             List.iter2_exn list1 list2 ~f:(fun x y ->
               Queue.enqueue q (f x y));
             Queue.to_list q)
            (let q = Queue.create () in
             List.iter pair_list ~f:(fun (x,y) ->
               Queue.enqueue q (f x y));
             Queue.to_list q))

    let%test_unit "rev_map2_exn vs rev + map2_exn" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * (t -> t -> t)]
        (G.tuple2
           (List.gen (G.tuple2 gen gen))
           (G.fn2 obs obs gen))
        ~f:(fun (pair_list, f) ->
          let list1, list2 = List.unzip pair_list in
          [%test_eq: t list]
            (List.rev_map2_exn list1 list2 ~f)
            (List.rev (List.map2_exn list1 list2 ~f)))

    let%test_unit "unzip + fold2_exn + fold" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * t * (t -> t -> t -> t)]
        (G.tuple3
           (List.gen (G.tuple2 gen gen))
           gen
           (G.fn3 obs obs obs gen))
        ~f:(fun (pair_list, init, f) ->
          let list1, list2 = List.unzip pair_list in
          [%test_eq: t]
            (List.fold2_exn list1 list2 ~init ~f)
            (List.fold (List.zip_exn list1 list2) ~init ~f:(fun acc (x,y) ->
               f acc x y)))

    let%test_unit "unzip + for_all2_exn vs for_all" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * (t -> t -> bool)]
        (G.tuple2
           (List.gen
              (G.tuple2 gen gen))
           (G.fn2 obs obs Bool.gen))
        ~f:(fun (pair_list, f) ->
          [%test_eq: bool]
            (let list1, list2 = List.unzip pair_list in
             List.for_all2_exn list1 list2 ~f)
            (List.for_all pair_list ~f:(fun (x,y) -> f x y)))

    let%test_unit "unzip + exists2_exn vs exists" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * (t -> t -> bool)]
        (G.tuple2
           (List.gen
              (G.tuple2 gen gen))
           (G.fn2 obs obs Bool.gen))
        ~f:(fun (pair_list, f) ->
          [%test_eq: bool]
            (let list1, list2 = List.unzip pair_list in
             List.exists2_exn list1 list2 ~f)
            (List.exists pair_list ~f:(fun (x,y) -> f x y)))

    let%test_unit "filter vs for_all" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_result: bool]
            (List.for_all ~f (List.filter ~f list))
            ~expect:true)

    let%test_unit "filter true" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_result: t list]
            (List.filter list ~f:(const true))
            ~expect:list)

    let%test_unit "filter vs rev_filter" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.rev (List.filter list ~f))
            (List.rev_filter list ~f))

    let%test_unit "filteri vs filter" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.filter list ~f)
            (List.filteri list ~f:(fun _ x -> f x)))

    let%test_unit "partition_map vs filter_map" =
      let partition_of_variant = function
        | `A a -> `Fst a
        | `B b -> `Snd b
      in
      Q.test ~sexp_of:[%sexp_of: t list * (t -> [ `Fst of t | `Snd of t ])]
        (G.tuple2 (List.gen gen)
           (G.fn
              obs
              (G.variant2 gen gen
               >>| partition_of_variant)))
        ~f:(fun (list, f) ->
          [%test_eq: t list * t list]
            (List.partition_map list ~f)
            (List.filter_map list ~f:(fun x ->
               match f x with
               | `Fst x -> Some x
               | `Snd _ -> None),
             List.filter_map list ~f:(fun x ->
               match f x with
               | `Fst _ -> None
               | `Snd x -> Some x)))

    let%test_unit "partition_tf vs partition_map" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list * t list]
            (List.partition_tf list ~f)
            (List.partition_map list ~f:(fun x ->
               if f x then `Fst x else `Snd x)))

    let%test_unit "append + split_n" =
      Q.test ~sexp_of:[%sexp_of: t list * t list]
        (G.tuple2
           (List.gen gen)
           (List.gen gen))
        ~f:(fun (list1, list2) ->
          [%test_result: t list * t list]
            (List.split_n (List.append list1 list2) (List.length list1))
            ~expect:(list1, list2))

    let%test_unit "sort vs stable_sort" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t -> int)]
        (G.tuple2
           (List.gen gen)
           (G.compare_fn obs))
        ~f:(fun (list, cmp) ->
          (* When comparing [t] using [cmp], [sort] and [stable_sort] should be
             indistinguishable. *)
          let compare = cmp in
          [%test_eq: t list]
            (List.sort        list ~cmp)
            (List.stable_sort list ~cmp))

    let%test_unit "stable_sort + merge vs append + stable_sort" =
      Q.test ~sexp_of:[%sexp_of: t list * t list * (t -> t -> int)]
        (G.tuple3
           (List.gen gen)
           (List.gen gen)
           (G.compare_fn obs))
        ~f:(fun (list1, list2, cmp) ->
          [%test_eq: t list]
            (List.merge ~cmp
               (List.stable_sort ~cmp list1)
               (List.stable_sort ~cmp list2))
            (List.stable_sort ~cmp
               (List.append list1 list2)))

    let%test_unit "hd vs hd_exn" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_eq: t option]
            (List.hd list)
            (Option.try_with (fun () -> List.hd_exn list)))

    let%test_unit "tl vs tl_exn" =
      Q.test ~sexp_of:[%sexp_of: t list]
        (List.gen gen)
        ~f:(fun list ->
          [%test_eq: t list option]
            (List.tl list)
            (Option.try_with (fun () -> List.tl_exn list)))

    let%test_unit "find vs find_exn" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t option]
            (List.find list ~f)
            (Option.try_with (fun () -> List.find_exn list ~f)))

    let%test_unit "find vs findi" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> bool)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs Bool.gen))
        ~f:(fun (list, f) ->
          [%test_eq: t option]
            (List.find list ~f)
            (Option.map ~f:snd (List.findi list ~f:(fun _ x -> f x))))

    let%test_unit "append + rev" =
      Q.test ~sexp_of:[%sexp_of: t list * t list]
        (G.tuple2
           (List.gen gen)
           (List.gen gen))
        ~f:(fun (list1, list2) ->
          [%test_eq: t list]
            (List.rev (List.append list1 list2))
            (List.append (List.rev list2) (List.rev list1)))

    let%test_unit "append associativity" =
      Q.test ~sexp_of:[%sexp_of: t list * t list * t list]
        (G.tuple3
           (List.gen gen)
           (List.gen gen)
           (List.gen gen))
        ~f:(fun (list1, list2, list3) ->
          [%test_eq: t list]
            (List.append list1 (List.append list2 list3))
            (List.append (List.append list1 list2) list3))

    let%test_unit "map + rev vs rev + map" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.rev (List.map list ~f))
            (List.map (List.rev list) ~f))

    let%test_unit "map + append vs append + map" =
      Q.test ~sexp_of:[%sexp_of: t list * t list * (t -> t)]
        (G.tuple3
           (List.gen gen)
           (List.gen gen)
           (G.fn obs gen))
        ~f:(fun (list1, list2, f) ->
          [%test_eq: t list]
            (List.append (List.map list1 ~f) (List.map list2 ~f))
            (List.map (List.append list1 list2) ~f))

    let%test_unit "map vs concat_map" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs gen))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.map list ~f)
            (List.concat_map list ~f:(fun x -> [ f x ])))

    let%test_unit "concat_mapi vs concat_map" =
      Q.test ~sexp_of:[%sexp_of: t list * (t -> t list)]
        (G.tuple2
           (List.gen gen)
           (G.fn obs (List.gen gen)))
        ~f:(fun (list, f) ->
          [%test_eq: t list]
            (List.concat_map list ~f)
            (List.concat_mapi list ~f:(fun _ x -> f x)))

    let%test_unit "unzip + map2_exn vs map" =
      Q.test ~sexp_of:[%sexp_of: (t * t) list * (t -> t -> t)]
        (G.tuple2
           (List.gen (G.tuple2 gen gen))
           (G.fn2 obs obs gen))
        ~f:(fun (pair_list, f) ->
          [%test_eq: t list]
            (let list1, list2 = List.unzip pair_list in
             List.map2_exn list1 list2 ~f)
            (List.map pair_list ~f:(fun (x,y) -> f x y)))

    let%test_unit "unzip + map3_exn vs map" =
      Q.test ~sexp_of:[%sexp_of: (t * (t * t)) list * (t -> t -> t -> t)]
        (G.tuple2
           (List.gen (G.tuple2 gen (G.tuple2 gen gen)))
           (G.fn3 obs obs obs gen))
        ~f:(fun (triple_list, f) ->
          [%test_eq: t list]
            (let list1, pair_list = List.unzip triple_list in
             let list2, list3     = List.unzip pair_list   in
             List.map3_exn list1 list2 list3 ~f)
            (List.map triple_list ~f:(fun (x,(y,z)) -> f x y z)))

    let%test_unit "rev + map3_exn vs rev_map3_exn" =
      Q.test ~sexp_of:[%sexp_of: (t * (t * t)) list * (t -> t -> t -> t)]
        (G.tuple2
           (List.gen (G.tuple2 gen (G.tuple2 gen gen)))
           (G.fn3 obs obs obs gen))
        ~f:(fun (triple_list, f) ->
          let list1, pair_list = List.unzip triple_list in
          let list2, list3     = List.unzip pair_list   in
          [%test_eq: t list]
            (List.rev_map3_exn list1 list2 list3 ~f)
            (List.rev (List.map3_exn list1 list2 list3 ~f)))

  end

  (* Float with bitwise comparison. *)
  module Float_ = struct

    include Float

    let compare x y =
      Int64.compare
        (Int64.bits_of_float x)
        (Int64.bits_of_float y)

    let equal x y =
      Int64.equal
        (Int64.bits_of_float x)
        (Int64.bits_of_float y)

  end

  let%test _ = Float_.equal Float.nan Float.nan

  module Bool'   = struct include Bool   let module_name = "Bool"   end
  module String' = struct include String let module_name = "String" end
  module Char'   = struct include Char   let module_name = "Char"   end
  module Int'    = struct include Int    let module_name = "Int"    end
  module Unit'   = struct include Unit   let module_name = "Unit"   end
  module Float'  = struct include Float_ let module_name = "Float"  end
  module Sexp'   = struct include Sexp   let module_name = "Sexp"   end

  let%test_module "unit w/ int"   = (module Make (Unit')   (Int'))
  let%test_module "bool w/ float" = (module Make (Bool')   (Float'))
  let%test_module "string w/ int" = (module Make (String') (Int'))
  let%test_module "char w/ float" = (module Make (Char')   (Float'))
  let%test_module "sexp w/ int"   = (module Make (Sexp')   (Int'))

  let%test_unit "List.fold_right doesn't allocate on empty lists" =
    let initial_words = Core_gc.minor_words () in
    let _ : int = List.fold_right [] ~f:(+) ~init:0 in
    let allocated = (Core_gc.minor_words ()) - initial_words in
    [%test_result: int] allocated ~expect:0

end)
